
--fixme: reorganize functions etc

--fixme: rename .reset functions to .redraw or .repaint or something
    -- reset is confusing since we also reset the game (and therefore reset the keyboard & guesses)

--fixme: standardize draw/hover/click definition order (picky detail)
--fixme: standardize assigning functions to tables
    -- table.key = function (arg) ...
    -- function table.key(arg) ...



-- word list module
local words = require("words")

-- localize common globals
--~ local getKeyFromScancode = love.keyboard.getKeyFromScancode
local lg = love.graphics
local math_random = math.random
local math_floor = math.floor
local math_ceil = math.ceil
local math_sin = math.sin

-- constants
local cloudMax = 6
local pi2 = math.pi * 2
local defFont = "unispace.regular.ttf"
local symFont = "NotoEmoji-Regular.ttf"
local qwerty = "qwertyuiopasdfghjklzxcvbnm" -- QWERTY layout
local color = {
    white    = {1, 1, 1, 1},
    black    = {0, 0, 0, 1},
    cloud    = {1, 1, 1, 0.3},
    hover    = {0.1, 0.1, 0.1},
    keyboard = {0.8, 0.8, 1, 0.6},
    control  = {0.4, 0.4, 0.7, 1},
    green    = {0.1, 0.5, 0.2, 0.9},
    yellow   = {1, 1, 0.1, 0.9},
    gray     = {0.1, 0.1, 0.0, 0.5},
    clear    = {1, 1, 1, 0}
}

-- audio
local music = love.audio.newSource("harp.ogg", "stream")
music:setLooping(true)
--fixme: add volumne control etc (settings panel)

-- gui objects
local gui = {}
local cell = {}
local chr = {}
local bkg = {}
local clouds = {}
local title = {}
local menu = {}
local help = {}
local pointer = {}
local guesses = {}
local keyboard = {}
local letters = {}
local backspace = {}
local clear = {}
local enter = {}


-- GRAPHICS CAPABILITIES (for cloud graphics)
local features = lg.getSupported()
if features.lighten then
    window.BlendMode, window.BlendAlphaMode = "lighten", "premultiplied" -- not supported on every system
else
    window.BlendMode, window.BlendAlphaMode = "add", "alphamultiply" -- works, but not as pretty
end
features = nil


-- DECLARE LOCAL FUNCTIONS

-- register gui item
local function guiRegister(item)
    gui[#gui + 1] = item
end

-- reset the game
local reset = {}
function reset.update()

    -- get new word
    words.getRndWord()

    -- reset guesses
    for i = 1, words.max do
        words[i] = {}
    end
    words.guess = 1

    -- reset keyboard colors
    for i = 1, #qwerty do
        letters[i].color = color.clear
    end

    -- reset guesses colors
    for guess = 1, words.max do
        for letter = 1, 5 do
            guesses[guess][letter] = color.clear
        end
    end

    -- reset visuals
    guesses.reset()
    keyboard.reset()
    pointer.reset()

    -- remove reset function from gui list
    gui[#gui] = nil
end

-- calculate largest font size possible for a rectangle
local function getFontSize(str, width, height, fontName)
    fontName = fontName or defFont
    local fontSize, fontObj, textObj = 0
    repeat
        if fontObj then fontObj:release(); textObj:release() end
        fontSize = fontSize + 1
        fontObj = lg.newFont(defFont, fontSize)
        textObj = lg.newText(fontObj, str)
    until textObj:getWidth() > width or textObj:getHeight() > height
    fontObj:release()
    textObj:release()
    return fontSize - 1
end

local function getLetterIndex(chr)
    return qwerty:find(chr:lower())
end

-- cloud init
local function resetCloud(cloud)

    -- init
    local size = math_floor(window.w / 5)

    -- set cloud specs
    cloud.scale = 1 / math_random(2, 3)
    cloud.speed = math_random(5, 15) * cloud.scale
    cloud.w = size --* math_random(1, 3) --fixme: improve cloud rendering
    cloud.h = size --* math_random(1, 2)
    cloud.x = math_random(0, window.w - cloud.w * cloud.scale)
    cloud.y = math_random(0, window.h - cloud.h * cloud.scale)

    -- draw a bunch of circles
    cloud.canvas = lg.newCanvas(cloud.w, cloud.h)
    lg.setCanvas(cloud.canvas)
    lg.setColor(color.cloud)
    local x, y, r
    local rMin = 5
    local rMax = math_floor(cloud.w / 3)
    for i = 1, 50 do
        r = math_random(rMin, rMax)
        x = math_random(r, cloud.w - r)
        y = math_random(r, cloud.h - r)
        lg.circle("fill", x, y, r)
    end
    lg.setCanvas()
    lg.setColor(color.white)
end

-- cloud init
local function newCloud()
    local cloud = {}
    cloud.draw = function ()
        lg.setBlendMode(window.BlendMode, window.BlendAlphaMode)
        lg.draw(cloud.canvas, cloud.x, cloud.y, 0, cloud.scale, cloud.scale * 0.5)
        lg.setBlendMode("alpha", "alphamultiply")
    end
    cloud.update = function (dt)
        cloud.x = cloud.x + dt * cloud.speed
        if cloud.x > window.x + window.w then
            resetCloud(cloud)
            cloud.x = -cloud.w * cloud.scale
        end
    end
    return cloud
end

-- draw a shadowed object
local function drawShadowed(drawable, x, y, offset) -- fixme: make shadowOffset dynamic based on resolution?
    lg.setColor(color.black)
    lg.draw(drawable, x + offset, y + offset)
    lg.setColor(color.white)
    lg.draw(drawable, x, y)
end

-- draw a shadowed glyph
local function drawGlyph(obj, shadowOffset, fontObj) -- fixme: make shadowOffset dynamic based on resolution?
    local x, y, w, h, chr, dx, dy = obj.x, obj.y, obj.w, obj.h, obj.chr
    if not fontObj then
        fontObj = cell.textFont
        dx = cell.textDx
        dy = cell.textDy
    else
        dx = cell.symDx
        dy = cell.symDy
    end
    if obj.color then
        lg.setColor(obj.color)
        lg.rectangle("fill", x, y, w, h, cell.r)
    end
    lg.setColor(color.black)
    lg.rectangle("line", x, y, w, h, cell.r)
    if chr then lg.print(chr, fontObj, x + dx + shadowOffset, y + dy + shadowOffset) end
    lg.setColor(color.white)
    lg.rectangle("line", x + 1, y + 1, w, h, cell.r)
    if chr then lg.print(chr, fontObj, x + dx, y + dy) end
end

-- basic hover function
local function hoverRect(rect)
    lg.setBlendMode("add")
    lg.setColor(color.hover)
    lg.rectangle("fill", rect.x, rect.y, rect.w, rect.h, cell.r)
    lg.setBlendMode("alpha")
    lg.setColor(color.white)
end

-- DEFINE GUI FUNCTIONS

function gui.letter(chr)
    --fixme: animate key presses/clicks
    if #words[words.guess] > 4 then return end
    local current = words[words.guess]
    current[#current + 1] = chr
    guesses.reset()
end
function gui.erase()
    local current = words[words.guess]
    current[#current] = nil
    guesses.reset()
end
function gui.clear()
    words[words.guess] = {}
    guesses.reset()
end
function gui.escape()
    if words[words.guess][1] then gui.clear() else menu.click() end
end
function gui.submit()

    -- validate submission
    if #words[words.guess] ~= 5 then return end --fixme: alert user
    local guess = table.concat(words[words.guess])
    if not words.isValid(guess) then return end --fixme: alert user

    -- init
    local answer = words.word
    local guessRow = guesses[words.guess]
    local letter

    -- assume all guesses are wrong
    for i = 1, 5 do
        guessRow[i] = color.gray
    end

    -- check for correct letters
    for i = 1, guess:len() do
        letter = guess:sub(i, i)
        if letter == answer:sub(i, i) then                    -- if the letter matches
            guessRow[i] = color.green                         -- color that guess green
            answer = answer:sub(1, i-1).."_"..answer:sub(i+1) -- and remove it from the word
            local key = letters[getLetterIndex(letter)]       -- identify keyboard letter
            key.color = color.green                           -- update keyboard color
        end
    end

    -- check for involved letters
    for i = 1, guess:len() do
        letter = guess:sub(i, i)
        if answer:find(letter) then                                           -- if the letter exists in the word
            if guessRow[i] ~= color.green then guessRow[i] = color.yellow end -- color that guess yellow
            answer = answer:gsub(letter, "_", 1)                              -- and remove it from the word
            local key = letters[getLetterIndex(letter)]                       -- identify keyboard letter
            if key.color ~= color.green then key.color = color.yellow end     -- update keyboard color
        end
    end

    -- check for (any remaining) wrong letters
    for i = 1, guess:len() do
        local letter = guess:sub(i, i)
        if letter ~= "_" then                                           -- for any remaining letters
            local key = letters[getLetterIndex(letter)]                 -- identify the keyboard letter
            if key.color == color.clear then key.color = color.gray end -- update keyboard color
        end
    end

    -- update visuals
    guesses.canvas:release()
    keyboard.canvas:release()
    guesses.reset() --fixme: animate the new letters?
    keyboard.reset() --fixme: animate/fade new colors?

    -- update game state
    if guess == words.word then
        state.current = state.win
        guiRegister(reset) -- new game when resuming this state
    else
        words.guess = words.guess + 1
        pointer.reset()
        if words.guess > words.max then
            state.current = state.lose
            guiRegister(reset) -- new game when resuming this state  --fixme: allow user to continue guessing before reset?
        end
    end
end


-- ONE-TIME INIT FOR GUI OBJECTS

-- background
bkg.x, bkg.y = 0, 0
bkg.draw = function ()
    lg.draw(bkg.img, -window.x, -window.y)
end
guiRegister(bkg)

-- clouds
math.randomseed(os.time())
for i = 1, cloudMax do
    clouds[i] = newCloud()
    guiRegister(clouds[i])
end

-- title
title.str = "Wordie"
title.draw = function () drawShadowed(title.textObj, title.x, title.y, 2) end
guiRegister(title)

-- menu
menu.draw = function () drawShadowed(menu.textObj, menu.x, menu.y, 2) end
function menu.hover()
    hoverRect(menu)
end
function menu.click()
    --~ print(words.word) --fixme: lol hack
    if music:isPlaying() then --fixme: lol hack
        music:pause()
    else
        music:play()
    end
end
guiRegister(menu)

-- help
help.draw = function () drawShadowed(help.textObj, help.x, help.y, 2) end
function help.hover()
    hoverRect(help)
end
function help.click()
    state.current = state.help
end
guiRegister(help)

-- pointer
pointer.chr = "»"
pointer.right = "«"
pointer.timer = 0
pointer.draw = function ()
    if state.current ~= state.game then return end
    drawShadowed(pointer.textObjL, pointer.x, pointer.y + pointer.dy, 2)
    drawShadowed(pointer.textObjR, guesses.x + guesses.w + cell.r, pointer.y + pointer.dy, 2)
end
pointer.reset = function ()
    pointer.x = guesses.x - math_floor(cell.w / 2)
    pointer.y = guesses.y + (words.guess - 1) * cell.h
end
pointer.update = function (dt)
    pointer.timer = pointer.timer + dt * 10
    if pointer.timer > pi2 then pointer.timer = pointer.timer - pi2 end
    pointer.dy = math_sin(pointer.timer)
end
guiRegister(pointer)

-- guesses
for guess = 1, words.max do
    guesses[guess] = {}
    for letter = 1, 5 do
        guesses[guess][letter] = color.clear
    end
end
function guesses.reset()
    local offset = offset or math_floor(math.max(window.w, window.h) * 0.0016) --fixme: test this!
    guesses.canvas = lg.newCanvas(guesses.w + cell.textDx, guesses.h + cell.textDy)
    lg.setCanvas(guesses.canvas)
    lg.clear()
    for guess = 1, words.max do
        for letter = 1, 5 do
            local x = (letter - 1) * cell.w + 1
            local y = (guess - 1) * cell.h + 1
            lg.setColor(guesses[guess][letter])
            lg.rectangle("fill", x, y, cell.w, cell.h, cell.r)
            lg.setColor(color.black)
            lg.rectangle("line", x, y, cell.w, cell.h, cell.r)
            if words[guess][letter] then lg.print(words[guess][letter], cell.textFont, x + cell.textDx + offset, y + cell.textDy + offset) end
            lg.setColor(color.white)
            lg.rectangle("line", x + 1, y + 1, cell.w, cell.h, cell.r)
            if words[guess][letter] then lg.print(words[guess][letter], cell.textFont, x + cell.textDx, y) end
        end
    end
    lg.setCanvas()
end
function guesses.draw()
    lg.draw(guesses.canvas, guesses.x, guesses.y)
end
guiRegister(guesses)

-- keyboard
function keyboard.reset()

    -- prep canvas
    keyboard.canvas = lg.newCanvas(keyboard.w, keyboard.h)
    lg.setCanvas(keyboard.canvas)
    lg.setColor(color.keyboard)
    lg.rectangle("fill", 0, 0, keyboard.w, keyboard.h, cell.r * 2)

    -- keyboard keys
    for i = 1, #qwerty do
        local key = letters[i]
        key.w = cell.w
        key.h = cell.h
        if i < 11 then -- top row
            key.x = cell.padding + (cell.w + cell.padding) * (i - 1)
            key.y = cell.padding
        elseif i < 20 then -- middle row
            key.x = cell.padding + (cell.w + cell.padding) * (i - 11)
            key.y = cell.padding * 2 + cell.h
        else -- bottom row
            key.x = cell.padding + (cell.w + cell.padding) * (i - 19) + math_floor(cell.w / 2)
            key.y = cell.padding * 3 + cell.h * 2
        end
        drawGlyph(key, 1)
        key.x = keyboard.x + key.x
        key.y = keyboard.y + key.y
    end

    -- special key: backspace
    backspace.w = cell.w
    backspace.h = cell.h
    backspace.x = (cell.w + cell.padding) * 9 + cell.padding
    backspace.y = cell.padding * 2 + cell.h
    drawGlyph(backspace, 1, cell.symFont)
    backspace.x = backspace.x + keyboard.x
    backspace.y = backspace.y + keyboard.y

    -- special key: clear
    clear.w = cell.w
    clear.h = cell.h
    clear.x = cell.padding
    clear.y = cell.padding * 3 + cell.h * 2
    drawGlyph(clear, 1, cell.symFont)
    clear.x = clear.x + keyboard.x
    clear.y = clear.y + keyboard.y

    -- special key: return
    enter.w = cell.w
    enter.h = cell.h
    enter.x = (cell.w + cell.padding) * 9 + cell.padding
    enter.y = cell.padding * 3 + cell.h * 2
    drawGlyph(enter, 1, cell.symFont)
    enter.x = enter.x + keyboard.x
    enter.y = enter.y + keyboard.y

    -- finish with keyboard
    lg.setCanvas()
    lg.setColor(color.white)
end
keyboard.draw = function () lg.draw(keyboard.canvas, keyboard.x, keyboard.y) end
guiRegister(keyboard)

-- basic letters
for i = 1, #qwerty do
    --~ local chr = getKeyFromScancode(qwerty:sub(i, i)):upper()
    local chr = qwerty:sub(i, i):upper() -- enforce US keys only, with QWERTY layout
    letters[i] = {chr=chr, color=color.clear}
    letters[i].click = function () gui.letter(chr) end
    letters[i].hover = function () hoverRect(letters[i]) end
    guiRegister(letters[i])
end

-- backspace key
backspace.hover = function () hoverRect(backspace) end
backspace.click = function () gui.erase() end
backspace.color = color.control
backspace.chr = '\xE2\x97\x80' -- left triangle
guiRegister(backspace)

-- clear key
clear.hover = function () hoverRect(clear) end
clear.click = function () gui.clear() end
clear.color = color.control
clear.chr = '\xE2\x8F\xAA' -- double left triangle
guiRegister(clear)

-- return key
enter.hover = function () hoverRect(enter) end
enter.click = function () gui.submit() end
enter.color = color.control
enter.chr = '\xE2\x9E\xA1' -- right triangle
guiRegister(enter)

-- (re)set graphics
function gui.resize(w, h)

    -- use safe area
    w, h = window.w, window.h -- confine to safe window size

    -- release old gui objects (anything with a defined .release method)
    for i = 1, #gui do
        for _, v in pairs(gui[i]) do
            if type(v) == "userdata" and v.release then v:release() end
        end
    end

    -- release non-gui objects
    if cell.textFont then
        cell.textFont:release()
        cell.symFont:release()
    end

    -- character cell size
    cell.padding = 5 --fixme: hardcode okay?
    cell.w = math_floor(w / 12)
    cell.h = math_floor(h / 13)
    cell.r = math_floor(math.max(cell.w, cell.h) * 0.05) --fixme: test & tweak
    cell.fontSize = getFontSize("#", cell.w, cell.h)

    -- text offsets
    cell.textFont = lg.newFont(defFont, cell.fontSize)
    local textObj = lg.newText(cell.textFont, "#")
    cell.textDx, cell.textDy = textObj:getDimensions()
    cell.textDx = math_floor((cell.w - cell.textDx) / 2)
    cell.textDy = math_floor((cell.h - cell.textDy) / 2)
    textObj:release()

    -- symbol offsets
    cell.symFont = lg.newFont(symFont, cell.fontSize)
    local textObj = lg.newText(cell.symFont, "#")
    cell.symDx, cell.symDy = textObj:getDimensions()
    cell.symDx = math_floor((cell.w - cell.symDx) / 2)
    cell.symDy = math_floor((cell.h - cell.symDy) / 2)
    textObj:release()

    -- adjust title
    title.y = math_floor(cell.h * 0.3)

    -- background gradient
    -- fixme: as guesses increase (especially with no yellow/green), redden the background
        -- use a less obvious curve than the one below
    bkg.w = w
    bkg.h = h
    bkg.ratio = 0.5 * (words.guess - 1) / (words.max - 1) + 1
    local ImageData = love.image.newImageData(lg.getDimensions())
    ImageData:mapPixel(function (_, y) -- Only y matters
        local ratio = y / (h - 1)
        local r = ratio * 0.55 + 0.45
        local g = ratio * 0.50 + 0.50 -- having slight differences between r and g makes for a smoother gradient
        --~ return r, g / bkg.ratio, 1 / bkg.ratio, 1
        return r, g, 1, 1
    end)
    bkg.img = lg.newImage(ImageData)
    ImageData:release()

    -- title
    local fontSize = getFontSize(title.str, w * 0.5, h * 0.15)
    local font = lg.newFont(defFont, fontSize)
    title.textObj = lg.newText(font, title.str)
    font:release()
    title.w, title.h = title.textObj:getDimensions()
    title.x = math_floor((w - title.w) / 2)

    -- menu
    menu.x = cell.w
    menu.y = cell.h
    menu.textObj = lg.newText(cell.symFont, '\xF0\x9F\x94\xA7') -- wrench Emoji symbol
    menu.w, menu.h = menu.textObj:getDimensions()

    -- help
    help.x = w - cell.w * 2
    help.y = cell.h
    help.textObj = lg.newText(cell.symFont, '\xE2\x9D\x93') -- Question mark Emoji symbol
    help.w, help.h = help.textObj:getDimensions()

    -- guesses
    guesses.w = cell.w * 5 + 2
    guesses.h = cell.h * words.max + 2
    guesses.x = math_floor((w - guesses.w) / 2)
    guesses.y = title.y + title.h + cell.textDx
    guesses.reset() --fixme: don't release canvas in redraw (just clear it instead)

    -- guess pointer
    pointer.textObjL = lg.newText(cell.textFont, pointer.chr)
    pointer.textObjR = lg.newText(cell.textFont, pointer.right)
    pointer.reset()

    -- keyboard
    keyboard.w = (cell.w + cell.padding) * 10 + cell.padding
    keyboard.h = (cell.h + cell.padding) * 3 + cell.padding
    keyboard.x = math_floor((w - keyboard.w) / 2)
    keyboard.y = guesses.y + cell.h * 6.5
    keyboard.reset()

    -- clouds
    for i = 1, #clouds do
        resetCloud(clouds[i])
    end
end

-- don't play game music if other music is already playing
if not love.system.hasBackgroundMusic() then music:play() end

return gui
