
local game = {}

local gui = require("game_load")

-- key bindings
local keybindings = {
    f1 = function () state.current = state.help end,
    backspace = gui.erase,
    escape = gui.escape,
    kpenter = gui.submit
}
keybindings["return"] = keybindings.kpenter -- reigster both enter keys
for i = 65, 90 do -- A to Z ASCII codes
    local chr = string.char(i)
    keybindings[chr] = function () gui.letter(chr) end
    keybindings[chr:lower()] = keybindings[chr] -- register upper & lower letter cases
end

local function isHovering(obj, x, y)
    return x >= obj.x + window.x and y >= obj.y + window.y and x <= obj.x + obj.w + window.x and y <= obj.y + obj.h + window.y
end

function game.draw()
    local x, y = love.mouse.getPosition()
    for i = 1, #gui do
        if gui[i].draw then gui[i].draw() end
        if state.current == state.game and gui[i].hover and isHovering(gui[i], x, y) then gui[i].hover(x, y) end
    end
end

function game.keyreleased(key)
    if keybindings[key] then keybindings[key]() end
end

function game.mousepressed(x, y)
    for i = #gui, 1, -1 do -- check objects top-down (shouldn't matter?)
        if gui[i].click and isHovering(gui[i], x, y) then
            gui.pressed = gui[i]
            return
        end
    end
end

function game.mousereleased(x, y)
    if not gui.pressed then return end
    if isHovering(gui.pressed, x, y) then gui.pressed.click() end
    gui.pressed = false
end

function game.resize(w, h)
    gui.resize(w, h)
end

function game.update(dt)
    if state.current ~= state.game then return end
    for i = 1, #gui do
        if gui[i].update then gui[i].update(dt) end
    end
end

return game
