
-- Wordie: an open source Wordle implementation
-- similar to https://www.nytimes.com/games/wordle/index.html
-- allows additional words to be completed too


-- TO DO
    -- sound effects (letters, submit, win, lose)
    -- music (https://soundimage.org/)
        -- add occasional bird song sounds?
    -- make Help & Lose states use window safe area
    -- save settings (window position, window size, music volume, sound volume)
    -- stats tracking
    -- new user instructions
    -- react to how user is doing
        -- score, based on guesses (fewer is better) and progress (yelows & greens)
        -- worse score affects game AV:
            -- reddened background color
            -- sound effects shift (ambience gets quieter, background tense strings get louder, etc)
    -- help >> MIT link >> fix link (itch?) once uploaded somewhere!
    -- improve win state
        -- fix/change win text
        -- winner's confetti
        -- provide a link to  https://en.wiktionary.org/wiki/[WORD]#English
            -- on win OR loss
        -- option to play again (random word)
    -- improve lose state
        -- option to keep guessing (still counts as a loss though)
    -- click "push button" effect for letters as they're activated (by mouse/touch/keyboard)
    -- add extra clouds added for yellow/green clues?
    -- Classic game allows 6 guesses
        -- maybe allow more after a "failure" for personal satisfaction
        -- track guesses and solve time and total puzzles done
    -- Settings
        -- sound toggle + volume control
        -- music toggle + volume control
        -- Game mode: Easy / Normal / Hard
            -- Easy mode: Unlimited guesses; stats not tracked
            -- Normal mode: default, rules as above
            -- Hard mode: Each guess MUST use all known clues; stats tracked separately from Normal (but you can still only score the daily ONCE)
            -- stats tracking: daily only, or also include additional games?
    -- On mobile (settings menu)
        -- add a portrait/landscape option
        -- add a quit button (not just mobile only?)
    -- make guesses list scrollable when more than 6 guesses present (Easy mode or post-fail guessing)
        -- up/down arrow keys
        -- mouse scroll wheel
        -- scroll bar?
    -- pipe dream:
        -- sync with current Wordle list & offset
        -- allow inputting letters in arbitrary order (mouse/arrow select cell, input a letter)?
        -- high scores / leaderboard (via https://dreamlo.com/ or similar)
        -- additional languages
        -- RTL support
        -- lang auto-detect, etc



--~ local profile = require("profile") -- debug stuff
--~ profile.start()


-- init window stuff
window = {sysFont = love.graphics.getFont()} -- store default font
window.isMobile = love.system.getOS() == "Android" or love.system.getOS() == "iOS" -- check for mobile device
if window.isMobile then
    love.window.setFullscreen(true)
    local w, h = love.graphics.getDimensions()
    if w > h then love.window.setMode(h, w) end -- initialize to portrait on mobile
end

-- state machine
state = {}
state.game = require("game_state")
state.help = require("help_state")
state.lose = require("lose_state")
state.win = require("win_state")
state.current = state.game

local love_graphics_translate = love.graphics.translate
function love.draw()
    love_graphics_translate(window.x, window.y)
    state.current.draw()
    love_graphics_translate(-window.x, -window.y) -- required or canvas redraw is affected (contrary to wiki)


    --~ print("KB used: "..collectgarbage("count")) -- fixme: slowly increases; why?

    --~ local stats = love.graphics.getStats()
    --~ for k,v in pairs(stats) do
        --~ print(k, v)
    --~ end

    --~ -- fixme: debug info
    --~ love.graphics.print("Current FPS: "..love.timer.getFPS(), 50, 150)
end

function love.keyreleased(key)
    if state.current.keyreleased then state.current.keyreleased(key) end
end

--~ function love.quit()
    --~ print(love.graphics.getDimensions())
    --~ print(profile.report())
--~ end

function love.mousepressed(x, y, button)
    if button ~= 1 then return end
    if state.current.mousepressed then state.current.mousepressed(x, y) end
end

function love.mousereleased(x, y, button)
    if button ~= 1 then return end
    if state.current.mousereleased then state.current.mousereleased(x, y) end
end

function love.touchpressed(_, x, y) -- touch press wrapper
    love.mousepressed(x, y, 1)
end

function love.touchreleased(_, x, y) -- touch release wrapper
    love.mousereleased(x, y, 1)
end

function love.resize(w, h) -- call any defined .resize functions (except "current" pseudo-state)
    window.x, window.y, window.w, window.h = love.window.getSafeArea() -- update safe area
    for k, v in pairs(state) do
        if v.resize and k ~= "current" then v.resize(w, h) end
    end
end

function love.update(dt)
    if state.current.update then state.current.update(dt) end
end

love.resize(love.graphics.getDimensions()) -- code & assets are all loaded; initialize graphics
